/* Copyright (C) 2022-2023 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Sergey Gorbunov, Sergei Zharko [committer] */

/// \file   CaUtils.h
/// \brief  Compile-time constants definition for the CA tracking algorithm
/// \since  02.06.2022
/// \author S.Zharko <s.zharko@gsi.de>

#pragma once  // include this header only once per compilation unit


#include "AlgoFairloggerCompat.h"
#include "CaConstants.h"
#include "CaSimd.h"

#include <sstream>

/// Namespace contains compile-time constants definition for the CA tracking algorithm
///
namespace cbm::algo::ca::utils
{

  inline fvec iif(const fmask& m, const fvec& t, const fvec& f) { return Vc::iif(m, t, f); }
  inline fvec fabs(const fvec& v) { return Vc::abs(v); }

  template<typename T>
  inline T iif(bool b, T t, T f)
  {
    return b ? t : f;
  }

  template<typename T>
  inline T fabs(const T& v)
  {
    return std::fabs(v);
  }

  /// tell the CPU that the bool condition is likely to be true
#if defined(__GNUC__) && __GNUC__ - 0 >= 3
  __attribute__((always_inline)) inline bool IsLikely(bool b) { return __builtin_expect(!!(b), 1); }
  __attribute__((always_inline)) inline bool IsUnlikely(bool b) { return __builtin_expect(!!(b), 0); }
#else
  inline bool IsLikely(bool b) { return b; }
  inline bool IsUnlikely(bool b) { return b; }
#endif

  /// \brief Checks whether a variable of a particular type defined
  /// \param val Value to be checked
  template<typename T>
  inline bool IsUndefined(T val)
  {
    if constexpr (std::is_same_v<T, float>) {
      return std::isnan(val);
    }
    else if constexpr (std::is_same_v<T, double>) {
      return std::isnan(val);
    }
    else if constexpr (std::is_same_v<T, fscal>) {
      return std::isnan(val);
    }
    else if constexpr (std::is_same_v<T, fvec>) {
      return isnan(val).isNotEmpty();
    }
    else {
      return val == cbm::algo::ca::constants::Undef<T>;
    }
  }

  /// \brief Checks whether a variable of a particular type is finite
  /// \param val  Value to check
  template<typename T>
  inline bool IsFinite(T val)
  {
    if constexpr (std::is_same_v<T, float>) {
      return std::isfinite(val);
    }
    else if constexpr (std::is_same_v<T, double>) {
      return std::isfinite(val);
    }
    else if constexpr (std::is_same_v<T, fscal>) {
      return std::isfinite(val);
    }
    else if constexpr (std::is_same_v<T, fvec>) {
      return isfinite(val).isFull();
    }
    else {
      return val != cbm::algo::ca::constants::Undef<T>;
    }
  }

  /// \brief Stingstream output operation for simd data
  template<typename DataT>
  void PrintSIMDmsg(std::stringstream& msg, DataT& v);

  /// \brief Checks, if a SIMD vector horizontally equal
  template<typename DataT>
  inline void CheckSimdVectorEquality(DataT /*v*/, const char* /*name*/)
  {
  }

  /// \brief Checks, if a SIMD vector horizontally equal
  /// TODO: Find this method in the VC!
  template<>
  void CheckSimdVectorEquality(fvec v, const char* name);

  /// \brief template based common fvec/float/double variable converter
  template<typename DataT, typename DataO>
  inline DataO GetValSimpl(const DataT& val)
  {
    return static_cast<DataO>(val);
  }

  /// \brief template based common fvec/float/double variable converter
  /// fvec to float implementation
  template<>
  inline float GetValSimpl(const fvec& val)
  {
    return val[0];
  }

  /// \brief template based common fvec/float/double variable converter
  /// fvec to double implementation
  template<>
  inline double GetValSimpl(const fvec& val)
  {
    return static_cast<double>(val[0]);
  }

  template<typename TdataA, typename TdataB, bool TDoAllA, bool TDoAllB>
  class VecCopySpec {
   public:
    [[gnu::always_inline]] static void CopyEntries(TdataA& a, int ia, const TdataB& b, int ib);
  };

  template<typename TdataA, typename TdataB>
  class VecCopySpec<TdataA, TdataB, true, true> {
   public:
    [[gnu::always_inline]] static void CopyEntries(TdataA& a, int, const TdataB& b, int) { a = b; }
  };

  template<typename TdataA, typename TdataB>
  class VecCopySpec<TdataA, TdataB, true, false> {
   public:
    [[gnu::always_inline]] static void CopyEntries(TdataA& a, int, const TdataB& b, int ib) { a = b[ib]; }
  };

  template<typename TdataA, typename TdataB>
  class VecCopySpec<TdataA, TdataB, false, true> {
   public:
    [[gnu::always_inline]] static void CopyEntries(TdataA& a, int ia, const TdataB& b, int) { a[ia] = b; }
  };

  template<typename TdataA, typename TdataB>
  class VecCopySpec<TdataA, TdataB, false, false> {
   public:
    [[gnu::always_inline]] static void CopyEntries(TdataA& a, int ia, const TdataB& b, int ib) { a[ia] = b[ib]; }
  };


  /// \brief Copies all/one SIMD entries from one class to the other class
  /// \details It helps to work with templates when classes might be SIMD or scalar
  /// \details If one of the classes is scalar, its SIMD element index is ignored
  ///
  /// \tparam TdataA  Type of the first class
  /// \tparam TdataB  Type of the second class
  /// \tparam TDoAllA  If true, all entries of the first class must be set
  /// \tparam TDoAllB  If true, all entries of the second class must be used
  /// \param a  First class
  /// \param ia  Index of SIMD vector element of the first class
  /// \param b  Second class
  /// \param ib  Index of SIMD vector element of the second class
  template<typename TdataA, typename TdataB, bool TDoAllA, bool TDoAllB>
  class VecCopy {
   public:
    static void CopyEntries(TdataA& a, int ia, const TdataB& b, int ib)
    {
      constexpr bool allA{TDoAllA || !std::is_same<TdataA, fvec>::value};
      constexpr bool allB{TDoAllB || !std::is_same<TdataB, fvec>::value};
      VecCopySpec<TdataA, TdataB, allA, allB>::CopyEntries(a, ia, b, ib);
    }
  };


  /// \brief  Checks, if a particular value lies within selected limits
  /// \param  name   Name of parameters
  /// \param  value  Value of parameter
  /// \param  lLimit Lower limit of parameter
  /// \param  uLimit Upper limit of parameter
  template<typename T>
  inline bool CheckValueLimits(const std::string& name, T value, T lLimit, T uLimit)
  {
    if (value < lLimit || value > uLimit) {
      LOG(error) << "parameter\033[1;32m" << name << "\033[0m (" << value << ") runs out of range: [" << lLimit << ','
                 << uLimit << ']';
      return false;
    }
    return true;
  }

  /// \struct EnumClassHash
  /// \brief  Hash for unordered_map with enum class keys
  struct EnumClassHash {
    template<typename T>
    int operator()(T t) const
    {
      return static_cast<int>(t);
    }
  };
}  // namespace cbm::algo::ca::utils
